// vi: set ts=4 sw=4 :
// vim: set tw=75 :

// meta_api.cpp - minimal implementation of metamod's plugin interface

// This is intended to illustrate the (more or less) bare minimum code
// required for a valid metamod plugin, and is targeted at those who want
// to port existing HL/SDK DLL code to run as a metamod plugin.

/*
 * Copyright (c) 2001-2003 Will Day <willday@hpgx.net>
 *
 *    This file is part of Metamod.
 *
 *    Metamod is free software; you can redistribute it and/or modify it
 *    under the terms of the GNU General Public License as published by the
 *    Free Software Foundation; either version 2 of the License, or (at
 *    your option) any later version.
 *
 *    Metamod is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with Metamod; if not, write to the Free Software Foundation,
 *    Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *    In addition, as a special exception, the author gives permission to
 *    link the code of this program with the Half-Life Game Engine ("HL
 *    Engine") and Modified Game Libraries ("MODs") developed by Valve,
 *    L.L.C ("Valve").  You must obey the GNU General Public License in all
 *    respects for all of the code used other than the HL Engine and MODs
 *    from Valve.  If you modify this file, you may extend this exception
 *    to your version of the file, but you are not obligated to do so.  If
 *    you do not wish to do so, delete this exception statement from your
 *    version.
 *
 */

#include <extdll.h>			// always

#include <meta_api.h>		// of course

#ifdef _WIN32
	#include "tlhelp32.h"
#endif

#ifdef LINUX
	#include <sys/mman.h>
	#ifndef DWORD
		typedef unsigned long DWORD;
	#endif
	#ifndef WORD
		typedef unsigned short WORD;
	#endif
	#ifndef PAGESIZE
		#define PAGESIZE 4096
	#endif
#endif

// Must provide at least one of these..
static META_FUNCTIONS gMetaFunctionTable = {
	NULL,				// pfnGetEntityAPI				HL SDK; called before game DLL
	NULL,				// pfnGetEntityAPI_Post			META; called after game DLL
	GetEntityAPI2,			// pfnGetEntityAPI2				HL SDK2; called before game DLL
	GetEntityAPI2_Post,			// pfnGetEntityAPI2_Post		META; called after game DLL
	NULL,			// pfnGetNewDLLFunctions		HL SDK2; called before game DLL
	NULL,			// pfnGetNewDLLFunctions_Post	META; called after game DLL
	NULL,			// pfnGetEngineFunctions	META; called before HL engine
	NULL,			// pfnGetEngineFunctions_Post	META; called after HL engine
};

#define MODULE_VERSION		"1.0"
#define MODULE_VERSION_F	1.0
#define MODULE_LOADABLE		PT_STARTUP
#define MODULE_UNLOADABLE	PT_NEVER

#define COMMAND_CONTINUE	0
#define COMMAND_HANDLED		1

// Description of plugin
plugin_info_t Plugin_info = {
	META_INTERFACE_VERSION,	// ifvers
	"UnicodePatch",	// name
	MODULE_VERSION,	// version
	"2010/10/16",	// date
	"Rulzy",	// author
	"http://www.dt-club.net/",	// url
	"UPATCH",	// logtag, all caps please
	MODULE_LOADABLE,	// (when) loadable
	MODULE_UNLOADABLE,	// (when) unloadable
};

cvar_t MODULE_CVAR_VERSION = {
	"upatch_version",
	MODULE_VERSION,
	FCVAR_SERVER,
	MODULE_VERSION_F,
	NULL
};

// Global vars from metamod:
meta_globals_t *gpMetaGlobals;		// metamod globals
gamedll_funcs_t *gpGamedllFuncs;	// gameDLL function tables
mutil_funcs_t *gpMetaUtilFuncs;		// metamod utility functions

struct CPlayer
{
	char Name[32];
	char Address[28];
	BOOL InGame;
} gPlayer[33];


char *FindMemory(char *Start, char *End, const char *Buffer, int BufferSize)
{
	for( char *p = Start; p <= End-BufferSize-1; p++ )
		if(!memcmp(p, Buffer, BufferSize))
			return p;
	return NULL;
}

struct CModuleInfo{
	void *BaseAddr;
	DWORD Size;
};

DLL_FUNCTIONS gFunctionTable;
C_DLLEXPORT	int	GetEntityAPI2(DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion)
{
	memset(&gFunctionTable, 0, sizeof(DLL_FUNCTIONS));
	memcpy(pFunctionTable, &gFunctionTable, sizeof(DLL_FUNCTIONS));
	return TRUE;
}

DLL_FUNCTIONS gFunctionTable_Post;
C_DLLEXPORT	int	GetEntityAPI2_Post(DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion)
{
	memset(&gFunctionTable_Post, 0, sizeof(DLL_FUNCTIONS));
	memcpy(pFunctionTable, &gFunctionTable_Post, sizeof(DLL_FUNCTIONS));
	return TRUE;
}

C_DLLEXPORT int Meta_Query(char *ifvers, plugin_info_t **pPlugInfo,	mutil_funcs_t *pMetaUtilFuncs) 
{
	*pPlugInfo = &Plugin_info;
	gpMetaUtilFuncs = pMetaUtilFuncs;

	int	mmajor = 0, mminor = 0,	pmajor = 0, pminor = 0;

	sscanf(ifvers, "%d:%d",	&mmajor, &mminor);
	sscanf(Plugin_info.ifvers, "%d:%d",	&pmajor, &pminor);

	if (strcmp(ifvers, Plugin_info.ifvers))
	{
		if (pmajor > mmajor)
		{
			LOG_ERROR(PLID, "metamod version is too old for this plugin; update metamod");
			return FALSE;
		} else if (pmajor < mmajor) {
			LOG_ERROR(PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
			return FALSE;
		} else if (pmajor == mmajor) {
			if (pminor > mminor) 
			{
				LOG_ERROR(PLID, "metamod version is too old for this plugin; update metamod");
				return FALSE;
			} else if (pminor < mminor) {
				LOG_MESSAGE(PLID, "warning: meta-interface version mismatch (plugin: %s, metamod: %s)", Plugin_info.ifvers, ifvers);
			}
		}
	}
	return TRUE;
}

#ifdef _WIN32
bool GetModuleInfo(const char *ModuleName, CModuleInfo *pMI )
{
	HANDLE hSnapshot = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, 0);
	if(hSnapshot == HANDLE(-1)) return false;
	MODULEENTRY32 ME;
	memset(&ME, 0, sizeof(MODULEENTRY32));
	ME.dwSize = sizeof(MODULEENTRY32);
	if(Module32First(hSnapshot, &ME))
	{
		if(!ModuleName || !*ModuleName)
		{
			pMI->BaseAddr = ME.modBaseAddr;
			pMI->Size = ME.modBaseSize;
			CloseHandle(hSnapshot);
			return true;
		}
		do
		{
			if(!strcasecmp(ME.szModule, ModuleName))
			{
				pMI->BaseAddr = ME.modBaseAddr;
				pMI->Size = ME.modBaseSize;
				CloseHandle(hSnapshot);
				return true;
			}
		} while (Module32Next(hSnapshot, &ME));
	}
	CloseHandle(hSnapshot);
	return false;
}


bool PatchSWDS(void *BaseAddr, DWORD BaseSize)
{
	// 035C68F3    33DB            xor     ebx, ebx
	// 035C68F5    8A1F            mov     bl, byte ptr [edi]
	// 035C68F7    47              inc     edi
	// 035C68F8    83FB 20         cmp     ebx, 20
	// 035C68FB    7C 28           jl      short 035C6925
	// 035C68FD    83FB 7E         cmp     ebx, 7E			==> nop; nop; nop;
	// 035C6900    7F 23           jg      short 035C6925	==> nop; nop;

	const char *Flag = "\x33\xDB\x8A\x1F\x47\x83\xFB\x20\x7C\x28\x83\xFB\x7E\x7F\x23";

	DWORD Oldp;
	BYTE *Offset;
	Offset=(BYTE*)FindMemory((char*)BaseAddr, (char*)BaseAddr+BaseSize-1, Flag, 0x0F);
	if(Offset)
	{
		VirtualProtect(Offset+0x0A, 5, PAGE_EXECUTE_READWRITE, &Oldp);
		// NOP .. NOP
		memset(Offset+0x0A, 0x90, 5);
		VirtualProtect(Offset+0x0A, 5, Oldp, &Oldp);
		return true;
	}
	return false;
}

bool PatchMP(void *BaseAddr, DWORD BaseSize)
{
	// 06D4A168  |>  8A06                  /mov     al, byte ptr [esi]
	// 06D4A16A  |.  84C0                  |test    al, al
	// 06D4A16C  |.  74 2A                 |je      short 06D4A198
	// 06D4A16E  |.  0FBEC0                |movsx   eax, al
	// 06D4A171  |.  50                    |push    eax
	// 06D4A172  |.  E8 XXXXXXXX           |call    XXXXXXXX
	
	// ==>

	// 06D4A168  |>  8A06                  /mov     al, byte ptr [esi]
	// 06D4A16A  |.  84C0                  |test    al, al
	// 06D4A16C  |.  74 2A                 |je      short 06D4A198
	// 06D4A16E  |.  3C 20                 |cmp     al, 20
	// 06D4A170  |.  76 04                 |jbe     short 06D4A176
	// 06D4A172  |.  B3 01                 |mov     bl, 1
	// 06D4A174  |.  EB 22                 |jmp     short 06D4A198
	// 06D4A176  |>  46                    |inc     esi
	// 06D4A177  |.^ 75 EF                 \jnz     short 06D4A168
	// 06D4A179  |.  90                    nop
	// ........  |.  90                    nop
	// 06D4A197  |.  90                    nop
	// 06D4A198  |>  84DB                  test    bl, bl
	// 06D4A19A  |.  0F84 61050000         je      06D4A701

	const char *Flag = "\x8A\x06\x84\xC0\x74\x2A\x0F\xBE\xC0\x50\xE8";


	DWORD Oldp;
	BYTE *Offset;
	Offset=(BYTE*)FindMemory((char*)BaseAddr, (char*)BaseAddr+BaseSize-1, Flag, 0x0B);
	if(Offset)
	{
		VirtualProtect(Offset, 30, PAGE_EXECUTE_READWRITE, &Oldp);
		*(WORD*)(Offset+0x06) = 0x203C;		// cmp     al, 20
		*(WORD*)(Offset+0x08) = 0x0476;		// jbe     short+4
		*(WORD*)(Offset+0x0A) = 0x01B3;		// mov     bl, 1
		*(WORD*)(Offset+0x0C) = 0x22EB;		// jmp     short +22
		*(Offset+0x0E) = 0x46;	           	// inc     esi
		*(WORD*)(Offset+0x0F) = 0xEF75;		// jnz     short -11
		memset(Offset+0x11, 0x90, 0x1F);	// nop ... nop
		VirtualProtect(Offset, 30, Oldp, &Oldp);
		return true;
	}
	return false;
}
#endif

#ifdef LINUX

char MBuffer[0x8000];

const char* ExtractFileName(const char *fpath) {
	int sl = strlen(fpath);
	const char *cp = fpath + sl;
	while ((DWORD)cp > (DWORD)fpath) {
		if (*cp == '\\' || *cp == '/') {
			return cp+1;
		}
		cp--;
	}
	return cp;
}

void* LocateLib(const char* libname)
{
	char fname[128], linebuf[512], clib[256];
	const char *clp;
	FILE *fl;
	int sl;
	void *RegStart, *RegEnd;
	Dl_info dli;

	sprintf(fname, "/proc/%d/maps", getpid());
	fl = fopen(fname, "r");
	if (fl == NULL) {
		return NULL;
	}

	setbuffer(fl, MBuffer, sizeof(MBuffer));
	while (fgets(linebuf, sizeof(linebuf), fl))
	{
		sl = sscanf(linebuf, "%x-%x %s %s %s %s %s", &RegStart, &RegEnd, fname, fname, fname, fname, clib);
		if (sl != 7)
			continue;

		if (dladdr(RegStart, &dli) == 0)
			continue;

		clp = ExtractFileName(dli.dli_fname);
		if (!strcmp(libname, clp)) {
			fclose(fl);
			return dli.dli_fbase;
		}
	}
	fclose(fl);
	return NULL;
}

DWORD GetFileSize(const char *FileName)
{
	FILE *fl = fopen(FileName, "rb");
	if (fl == NULL)	return 0;
	int FileSize;
	fseek(fl, 0, SEEK_END);
	FileSize = ftell(fl);
	fseek(fl, 0, SEEK_SET);	
	fclose(fl);
	if (FileSize<0)
		return 0;
	return (DWORD)FileSize;
}

bool GetModuleInfo(const char *ModulePath, CModuleInfo *pMI )
{
	const char *ModuleName = ExtractFileName(ModulePath);
	void *lib = LocateLib(ModuleName);
	if (lib == NULL)
			return false;

	DWORD FileSize = GetFileSize(ModulePath);
	if( FileSize==0 )
		return false;
	pMI->BaseAddr = lib;	
	pMI->Size = FileSize;
	return true;
}

bool MProtect_Ex(void *addr, int npages) {
	void *paddr;
	paddr = (void *)(((size_t)addr) & ~(PAGESIZE-1));
	return !mprotect(paddr, PAGESIZE*(npages+1), PROT_READ | PROT_WRITE | PROT_EXEC);
}

bool PatchEngine(void *BaseAddr, DWORD BaseSize)
{
	// 00048150 0FB637                  movzx esi, byte ptr [edi]
	// 00048153 47                      inc edi
	// 00048154 8D46E0                  lea eax, dword ptr [esi-20]		==> 909090	nop; nop; nop;
	// 00048157 83F85E                  cmp eax, 0000005E			==> 83FE20	cmp esi, 20
	// 0004815A 77XX                    ja XXXXXXXX				==> 72XX	jb XXXXXXXX

	const char *Flag = "\x0F\xB6\x37\x47\x8D\x46\xE0\x83\xF8\x5E\x77";
	const char *Repl = "\x0F\xB6\x37\x47\x90\x90\x90\x83\xFE\x20\x72";

	BYTE *Offset;
	Offset=(BYTE*)FindMemory((char*)BaseAddr, (char*)BaseAddr+BaseSize-1, Flag, 0x0B);
	if(Offset)
	{
		MProtect_Ex(Offset, 1);
		memmove(Offset, Repl, 0x0B);
		return true;
	}
	return false;
}

bool PatchCsI386(void *BaseAddr, DWORD BaseSize)
{
	// 000902E7 8B8D 68FEFFFF           mov ecx, dword ptr [ebp-198]
	// 000902ED C685 57FEFFFF 00        mov byte ptr [ebp-1A9], 00
	// 000902F4 85C9                    test ecx, ecx
	// 000902F6 74 2E                   je 00090326
	// 000902F8 803900                  cmp byte ptr [ecx], 00
	// 000902FB 74 29                   je 00090326
	// 000902FD 8B93 402C0000           mov edx, dword ptr [ebx+00002C40]	// Flag begin
	// 00090303 0FBE01                  movsx eax, byte ptr [ecx]			
	// 00090306 8B12                    mov edx, dword ptr [edx]
	// 00090308 0FB70442                movzx eax, word ptr [edx+2*eax]
	// 0009030C F6C440                  test ah, 40							
	// 0009030F 740F                    je 00090320
	// 00090311 F6C420                  test ah, 20
	// 00090314 750D                    jne 00090323
	// 00090316 C68557FEFFFF01          mov byte ptr [ebp-1A9], 01
	// 0009031D EB04                    jmp 00090323
	// 0009031F 90                      nop
	// 00090320 C60120                  mov byte ptr [ecx], 20
	// 00090323 41                      inc ecx
	// 00090324 75D2                    jne 000902F8						// Flag end
	// 00090326 80BD57FEFFFF00          cmp byte ptr [ebp-1A9], 00
	// 0009032D 0F8499080000            je 00090BCC

	// ==>

	// 000902E7      8B8D 68FEFFFF         mov     ecx, dword ptr [ebp-198]
	// 000902ED      C685 57FEFFFF 00      mov     byte ptr [ebp-1A9], 0
	// 000902F4      85C9                  test    ecx, ecx
	// 000902F6      74 2E                 je      short 00090326
	// 000902F8      8039 00               cmp     byte ptr [ecx], 0
	// 000902FB      74 29                 je      short 00090326
	// 000902FD      8039 20               cmp     byte ptr [ecx], 20
	// 00090300      76 09                 jbe     short 0009030B
	// 00090302      C685 57FEFFFF 01      mov     byte ptr [ebp-1A9], 1
	// 00090309      EB 1B                 jmp     short 00090326
	// 0009030B      41                    inc     ecx
	// 0009030C    ^ 75 EA                 jnz     short 000902F8
	// 0009030E      90                    nop
	// ........      90                    nop
	// 00090325      90                    nop
	// 00090326      80BD 57FEFFFF 00      cmp     byte ptr [ebp-1A9], 0
	// 0009032D      0F84 99080000         je      0144DEB5


	const char *Flag = "\x8B\x93\x40\x2C\x00\x00\x0F\xBE\x01\x8B\x12\x0F\xB7\x04\x42\xF6\xC4\x40\x74\x0F\xF6\xC4\x20\x75\x0D\xC6\x85\x57\xFE\xFF\xFF\x01\xEB\x04\x90\xC6\x01\x20\x41\x75\xD2";
	const char *Repl = "\x80\x39\x20\x76\x09\xC6\x85\x57\xFE\xFF\xFF\x01\xEB\x1B\x41\x75\xEA\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90\x90";

	BYTE *Offset;
	Offset=(BYTE*)FindMemory((char*)BaseAddr, (char*)BaseAddr+BaseSize-1, Flag, 0x29);
	if(Offset)
	{
		MProtect_Ex(Offset, 1);
		memmove(Offset, Repl, 0x29);
		return true;
	}
	return false;
}
#endif

char* trimbuf(char *str) {
 	char *ibuf = str;
	int i = 0;

	if (str == NULL) return NULL;
	for (ibuf = str; *ibuf && isspace(*ibuf); ++ibuf)
		;
	
	i = strlen(ibuf);
	if (str != ibuf)
		memmove(str, ibuf, i);

	str[i] = 0;
	while (--i >= 0) {
		if (!isspace(ibuf[i]))
			break;
	}
	ibuf[++i] = 0;
	return str;
}

int PatchUnicodePlayerName = 1, PatchUnicodeChat = 1;

void LoadCfg() {
	const char *UPATCH_CFG = "upatch.cfg";
	char buf[1024];
	char gamedir[256];
	char *cl;
	char *vl;
	int cline = 0;

	sprintf(buf, "./%s", UPATCH_CFG);
	FILE *fl = fopen(buf, "r");
	if (fl == NULL) {
		GET_GAME_DIR(gamedir);
		sprintf(buf, "./%s/%s", gamedir, UPATCH_CFG);
		fl = fopen(buf, "r");
		if (fl == NULL) {
			return;
		}
	}

	while (fgets(buf, sizeof(buf), fl)) {
		cline++;
		cl = trimbuf(buf);
		if (strlen(cl) == 0 || cl[0] == '#' || (cl[0] == '/' && cl[1] == '/') ) continue;
		vl = strchr(cl, '=');
		if (vl == NULL) {
			continue;
		}
		*(vl++) = 0;
		cl = trimbuf(cl);
		vl = trimbuf(vl);
		if ( !strcasecmp(cl, "PatchUnicodePlayerName") )
		{
			PatchUnicodePlayerName = atoi(vl);
		} else if ( !strcasecmp(cl, "PatchUnicodeChat") ) {
			PatchUnicodeChat = atoi(vl);
		}
	}
	fclose(fl);
	return;
}

C_DLLEXPORT int Meta_Attach(PLUG_LOADTIME  now, META_FUNCTIONS *pFunctionTable,
					meta_globals_t *pMGlobals, gamedll_funcs_t *pGamedllFuncs) 
{
	if(now>MODULE_LOADABLE)
	{
		LOG_ERROR(PLID, "Can't load plugin right now");
		return FALSE;
	}
	if(!pMGlobals) {
		LOG_ERROR(PLID, "Meta_Attach called with null pMGlobals");
		return FALSE;
	}
	gpMetaGlobals=pMGlobals;
	if(!pFunctionTable) {
		LOG_ERROR(PLID, "Meta_Attach called with null pFunctionTable");
		return FALSE;
	}
	memcpy(pFunctionTable, &gMetaFunctionTable, sizeof(META_FUNCTIONS));
	gpGamedllFuncs=pGamedllFuncs;

	CVAR_REGISTER(&MODULE_CVAR_VERSION);

	SERVER_PRINT("\n");

	LoadCfg();

	CModuleInfo MI;
	bool bSuccess = false;

	if( PatchUnicodePlayerName )
	{

#ifdef _WIN32
		if(GetModuleInfo("swds.dll", &MI))
			bSuccess = PatchSWDS(MI.BaseAddr, MI.Size);
#endif

#ifdef LINUX
		if ( GetModuleInfo("./engine_amd.so", &MI) || GetModuleInfo("./engine_i486.so", &MI) || GetModuleInfo("./engine_i686.so", &MI) )
			bSuccess = PatchEngine(MI.BaseAddr, MI.Size);
#endif
		if( bSuccess )
			SERVER_PRINT("[UPATCH] Patch \"Unicode Player Name\" successful!\n");
		else
			SERVER_PRINT("[UPATCH] Patch \"Unicode Player Name\" failed.\n");
	}
	

	if( PatchUnicodeChat )
	{
		bSuccess = false;
#ifdef _WIN32
		if(GetModuleInfo("mp.dll", &MI))
			bSuccess = PatchMP(MI.BaseAddr, MI.Size);
#endif

#ifdef LINUX
		if ( GetModuleInfo("./cstrike/dlls/cs_i386.so", &MI) )
			bSuccess = PatchCsI386(MI.BaseAddr, MI.Size);
#endif

		if( bSuccess )
			SERVER_PRINT("[UPATCH] Patch \"Unicode Chat\" successful!\n");
		else
			SERVER_PRINT("[UPATCH] Patch \"Unicode Chat\" failed.\n");
	}

	SERVER_PRINT("\n");
	return TRUE;
}

C_DLLEXPORT int Meta_Detach(PLUG_LOADTIME now, PL_UNLOAD_REASON reason) 
{
	if (now>MODULE_UNLOADABLE && reason!=PNL_CMD_FORCED)
	{
		LOG_ERROR(PLID, "Can't unload plugin right now");
		return FALSE;
	}
	return TRUE;
}
